@@ -138,7 +138,8 @@ Downstream::Downstream(Upstream *upstream, MemchunkPool *mcpool,
       chunked_response_(false),
       expect_final_response_(false),
       request_pending_(false),
-      request_header_sent_(false) {
+      request_header_sent_(false),
+      accesslog_written_(false) {
 
   auto &timeoutconf = get_config()->http2.timeout;
 
@@ -350,10 +351,8 @@ void Downstream::crumble_request_cookie(std::vector<nghttp2_nv> &nva) {
 }
 
 namespace {
-void add_header(bool &key_prev, size_t &sum, HeaderRefs &headers,
-                const StringRef &name, const StringRef &value, bool no_index,
-                int32_t token) {
-  key_prev = true;
+void add_header(size_t &sum, HeaderRefs &headers, const StringRef &name,
+                const StringRef &value, bool no_index, int32_t token) {
   sum += name.size() + value.size();
   headers.emplace_back(name, value, no_index, token);
 }
@@ -446,14 +445,14 @@ const HeaderRefs::value_type *FieldStore::header(const StringRef &name) const {
 
 void FieldStore::add_header_token(const StringRef &name, const StringRef &value,
                                   bool no_index, int32_t token) {
-  shrpx::add_header(header_key_prev_, buffer_size_, headers_, name, value,
-                    no_index, token);
+  shrpx::add_header(buffer_size_, headers_, name, value, no_index, token);
 }
 
 void FieldStore::alloc_add_header_name(const StringRef &name) {
   auto name_ref = alloc_header_name(balloc_, name);
   auto token = http2::lookup_token(name_ref);
   add_header_token(name_ref, StringRef{}, false, token);
+  header_key_prev_ = true;
 }
 
 void FieldStore::append_last_header_key(const char *data, size_t len) {
@@ -466,21 +465,24 @@ void FieldStore::append_last_header_value(const char *data, size_t len) {
                                   headers_, data, len);
 }
 
-void FieldStore::clear_headers() { headers_.clear(); }
+void FieldStore::clear_headers() {
+  headers_.clear();
+  header_key_prev_ = false;
+}
 
 void FieldStore::add_trailer_token(const StringRef &name,
                                    const StringRef &value, bool no_index,
                                    int32_t token) {
   // Header size limit should be applied to all header and trailer
   // fields combined.
-  shrpx::add_header(trailer_key_prev_, buffer_size_, trailers_, name, value,
-                    no_index, token);
+  shrpx::add_header(buffer_size_, trailers_, name, value, no_index, token);
 }
 
 void FieldStore::alloc_add_trailer_name(const StringRef &name) {
   auto name_ref = alloc_header_name(balloc_, name);
   auto token = http2::lookup_token(name_ref);
   add_trailer_token(name_ref, StringRef{}, false, token);
+  trailer_key_prev_ = true;
 }
 
 void FieldStore::append_last_trailer_key(const char *data, size_t len) {
@@ -905,7 +907,9 @@ void Downstream::disable_downstream_wtimer() {
   disable_timer(loop, &downstream_wtimer_);
 }
 
-bool Downstream::accesslog_ready() const { return resp_.http_status > 0; }
+bool Downstream::accesslog_ready() const {
+  return !accesslog_written_ && resp_.http_status > 0;
+}
 
 void Downstream::add_retry() { ++num_retry_; }
 
@@ -921,6 +925,10 @@ bool Downstream::get_request_pending() const { return request_pending_; }
 
 void Downstream::set_request_header_sent(bool f) { request_header_sent_ = f; }
 
+bool Downstream::get_request_header_sent() const {
+  return request_header_sent_;
+}
+
 bool Downstream::request_submission_ready() const {
   return (request_state_ == Downstream::HEADER_COMPLETE ||
           request_state_ == Downstream::MSG_COMPLETE) &&
@@ -980,4 +988,6 @@ void Downstream::set_addr(const DownstreamAddr *addr) { addr_ = addr; }
 
 const DownstreamAddr *Downstream::get_addr() const { return addr_; }
 
+void Downstream::set_accesslog_written(bool f) { accesslog_written_ = f; }
+
 } // namespace shrpx
