@@ -48,6 +48,10 @@ void timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
 
+  if (w == &conn->rt && !conn->expired_rt()) {
+    return;
+  }
+
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, dconn) << "Time out";
   }
@@ -71,11 +75,12 @@ void connect_timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
   auto addr = dconn->get_addr();
+  auto raddr = dconn->get_raddr();
 
   DCLOG(WARN, dconn) << "Connect time out; addr="
-                     << util::to_numeric_addr(&addr->addr);
+                     << util::to_numeric_addr(raddr);
 
-  downstream_failure(addr);
+  downstream_failure(addr, raddr);
 
   auto downstream = dconn->get_downstream();
   auto upstream = downstream->get_upstream();
@@ -112,15 +117,51 @@ void readcb(struct ev_loop *loop, ev_io *w, int revents) {
 }
 } // namespace
 
+namespace {
+void backend_retry(Downstream *downstream) {
+  auto upstream = downstream->get_upstream();
+  auto handler = upstream->get_client_handler();
+
+  downstream->add_retry();
+
+  if (downstream->no_more_retry()) {
+    delete handler;
+    return;
+  }
+
+  downstream->pop_downstream_connection();
+
+  auto ndconn = handler->get_downstream_connection(downstream);
+  if (ndconn) {
+    if (downstream->attach_downstream_connection(std::move(ndconn)) == 0) {
+      return;
+    }
+  }
+
+  downstream->set_request_state(Downstream::CONNECT_FAIL);
+
+  if (upstream->on_downstream_abort_request(downstream, 503) != 0) {
+    delete handler;
+  }
+}
+} // namespace
+
 namespace {
 void writecb(struct ev_loop *loop, ev_io *w, int revents) {
+  int rv;
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
   auto downstream = dconn->get_downstream();
   auto upstream = downstream->get_upstream();
   auto handler = upstream->get_client_handler();
 
-  if (upstream->downstream_write(dconn) != 0) {
+  rv = upstream->downstream_write(dconn);
+  if (rv == SHRPX_ERR_RETRY) {
+    backend_retry(downstream);
+    return;
+  }
+
+  if (rv != 0) {
     delete handler;
   }
 }
@@ -131,23 +172,8 @@ void connectcb(struct ev_loop *loop, ev_io *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
   auto downstream = dconn->get_downstream();
-  auto upstream = downstream->get_upstream();
-  auto handler = upstream->get_client_handler();
   if (dconn->connected() != 0) {
-    downstream->pop_downstream_connection();
-
-    auto ndconn = handler->get_downstream_connection(downstream);
-    if (ndconn) {
-      if (downstream->attach_downstream_connection(std::move(ndconn)) == 0) {
-        return;
-      }
-    }
-
-    downstream->set_request_state(Downstream::CONNECT_FAIL);
-
-    if (upstream->on_downstream_abort_request(downstream, 503) != 0) {
-      delete handler;
-    }
+    backend_retry(downstream);
     return;
   }
   writecb(loop, w, revents);
@@ -163,28 +189,52 @@ HttpDownstreamConnection::HttpDownstreamConnection(
             readcb, connect_timeoutcb, this,
             get_config()->tls.dyn_rec.warmup_threshold,
             get_config()->tls.dyn_rec.idle_timeout, PROTO_HTTP1),
-      do_read_(&HttpDownstreamConnection::noop),
-      do_write_(&HttpDownstreamConnection::noop),
-      do_signal_write_(&HttpDownstreamConnection::noop),
+      on_read_(&HttpDownstreamConnection::noop),
+      on_write_(&HttpDownstreamConnection::noop),
+      signal_write_(&HttpDownstreamConnection::noop),
       worker_(worker),
       ssl_ctx_(worker->get_cl_ssl_ctx()),
       group_(group),
       addr_(nullptr),
+      raddr_(nullptr),
       ioctrl_(&conn_.rlimit),
       response_htp_{0},
-      initial_addr_idx_(initial_addr_idx) {}
+      initial_addr_idx_(initial_addr_idx),
+      reuse_first_write_done_(true),
+      reusable_(true) {}
 
 HttpDownstreamConnection::~HttpDownstreamConnection() {
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, this) << "Deleted";
   }
+
+  if (dns_query_) {
+    auto dns_tracker = worker_->get_dns_tracker();
+    dns_tracker->cancel(dns_query_.get());
+  }
 }
 
 int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
+  int rv;
+
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, this) << "Attaching to DOWNSTREAM:" << downstream;
   }
 
+  downstream_ = downstream;
+
+  rv = initiate_connection();
+  if (rv != 0) {
+    downstream_ = nullptr;
+    return rv;
+  }
+
+  return 0;
+}
+
+int HttpDownstreamConnection::initiate_connection() {
+  int rv;
+
   auto worker_blocker = worker_->get_connect_blocker();
   if (worker_blocker->blocked()) {
     if (LOG_ENABLED(INFO)) {
@@ -208,42 +258,117 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
         shared_addr->affinity == AFFINITY_NONE ? shared_addr->next : temp_idx;
     auto end = next_downstream;
     for (;;) {
-      auto &addr = addrs[next_downstream];
+      auto check_dns_result = dns_query_.get() != nullptr;
 
-      if (++next_downstream >= addrs.size()) {
-        next_downstream = 0;
-      }
+      DownstreamAddr *addr;
+      if (check_dns_result) {
+        addr = addr_;
+        addr_ = nullptr;
+        assert(addr);
+        assert(addr->dns);
+      } else {
+        assert(addr_ == nullptr);
+        addr = &addrs[next_downstream];
 
-      if (addr.proto != PROTO_HTTP1) {
-        if (end == next_downstream) {
-          return SHRPX_ERR_NETWORK;
+        if (++next_downstream >= addrs.size()) {
+          next_downstream = 0;
         }
 
-        continue;
+        if (addr->proto != PROTO_HTTP1) {
+          if (end == next_downstream) {
+            return SHRPX_ERR_NETWORK;
+          }
+
+          continue;
+        }
       }
 
-      auto &connect_blocker = addr.connect_blocker;
+      auto &connect_blocker = addr->connect_blocker;
 
       if (connect_blocker->blocked()) {
         if (LOG_ENABLED(INFO)) {
-          DCLOG(INFO, this) << "Backend server "
-                            << util::to_numeric_addr(&addr.addr)
-                            << " was not available temporarily";
+          DCLOG(INFO, this) << "Backend server " << addr->host << ":"
+                            << addr->port << " was not available temporarily";
         }
 
-        if (end == next_downstream) {
+        if (check_dns_result) {
+          dns_query_.reset();
+        } else if (end == next_downstream) {
           return SHRPX_ERR_NETWORK;
         }
 
         continue;
       }
 
-      conn_.fd = util::create_nonblock_socket(addr.addr.su.storage.ss_family);
+      Address *raddr;
+
+      if (addr->dns) {
+        if (!check_dns_result) {
+          auto dns_query = make_unique<DNSQuery>(
+              addr->host, [this](int status, const Address *result) {
+                int rv;
+
+                if (status == DNS_STATUS_OK) {
+                  *this->resolved_addr_ = *result;
+                }
+
+                rv = this->initiate_connection();
+                if (rv != 0) {
+                  // This callback destroys |this|.
+                  auto downstream = this->downstream_;
+                  backend_retry(downstream);
+                }
+              });
+
+          auto dns_tracker = worker_->get_dns_tracker();
+
+          if (!resolved_addr_) {
+            resolved_addr_ = make_unique<Address>();
+          }
+          rv = dns_tracker->resolve(resolved_addr_.get(), dns_query.get());
+          switch (rv) {
+          case DNS_STATUS_ERROR:
+            downstream_failure(addr, nullptr);
+            if (end == next_downstream) {
+              return SHRPX_ERR_NETWORK;
+            }
+            continue;
+          case DNS_STATUS_RUNNING:
+            dns_query_ = std::move(dns_query);
+            // Remember current addr
+            addr_ = addr;
+            return 0;
+          case DNS_STATUS_OK:
+            break;
+          default:
+            assert(0);
+          }
+        } else {
+          switch (dns_query_->status) {
+          case DNS_STATUS_ERROR:
+            dns_query_.reset();
+            downstream_failure(addr, nullptr);
+            continue;
+          case DNS_STATUS_OK:
+            dns_query_.reset();
+            break;
+          default:
+            assert(0);
+          }
+        }
+
+        raddr = resolved_addr_.get();
+        util::set_port(*resolved_addr_, addr->port);
+      } else {
+        raddr = &addr->addr;
+      }
+
+      conn_.fd = util::create_nonblock_socket(raddr->su.storage.ss_family);
 
       if (conn_.fd == -1) {
         auto error = errno;
         DCLOG(WARN, this) << "socket() failed; addr="
-                          << util::to_numeric_addr(&addr.addr)
+                          << util::to_numeric_addr(raddr)
                           << ", errno=" << error;
 
         worker_blocker->on_failure();
@@ -253,20 +378,19 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
 
       worker_blocker->on_success();
 
-      int rv;
-      rv = connect(conn_.fd, &addr.addr.su.sa, addr.addr.len);
+      rv = connect(conn_.fd, &raddr->su.sa, raddr->len);
       if (rv != 0 && errno != EINPROGRESS) {
         auto error = errno;
         DCLOG(WARN, this) << "connect() failed; addr="
-                          << util::to_numeric_addr(&addr.addr)
+                          << util::to_numeric_addr(raddr)
                           << ", errno=" << error;
 
-        downstream_failure(&addr);
+        downstream_failure(addr, raddr);
 
         close(conn_.fd);
         conn_.fd = -1;
 
-        if (end == next_downstream) {
+        if (!check_dns_result && end == next_downstream) {
           return SHRPX_ERR_NETWORK;
         }
 
@@ -278,7 +402,8 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
         DCLOG(INFO, this) << "Connecting to downstream server";
       }
 
-      addr_ = &addr;
+      addr_ = addr;
+      raddr_ = raddr;
 
       if (addr_->tls) {
         assert(ssl_ctx_);
@@ -319,14 +444,19 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
     ev_timer_again(conn_.loop, &conn_.wt);
   } else {
     // we may set read timer cb to idle_timeoutcb.  Reset again.
-    conn_.rt.repeat = downstreamconf.timeout.read;
     ev_set_cb(&conn_.rt, timeoutcb);
-    ev_timer_stop(conn_.loop, &conn_.rt);
+    if (conn_.read_timeout < downstreamconf.timeout.read) {
+      conn_.read_timeout = downstreamconf.timeout.read;
+      conn_.last_read = ev_now(conn_.loop);
+    } else {
+      conn_.again_rt(downstreamconf.timeout.read);
+    }
 
     ev_set_cb(&conn_.rev, readcb);
-  }
 
-  downstream_ = downstream;
+    on_write_ = &HttpDownstreamConnection::write_reuse_first;
+    reuse_first_write_done_ = false;
+  }
 
   http_parser_init(&response_htp_, HTTP_RESPONSE);
   response_htp_.data = downstream_;
@@ -335,6 +465,10 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
 }
 
 int HttpDownstreamConnection::push_request_headers() {
+  if (downstream_->get_request_header_sent()) {
+    return 0;
+  }
+
   const auto &downstream_hostport = addr_->hostport;
   const auto &req = downstream_->request();
 
@@ -366,7 +500,7 @@ int HttpDownstreamConnection::push_request_headers() {
   // Assume that method and request path do not contain \r\n.
   auto meth = http2::to_method_string(req.method);
   buf->append(meth);
-  buf->append(" ");
+  buf->append(' ');
 
   if (connect_method) {
     buf->append(authority);
@@ -561,6 +695,8 @@ int HttpDownstreamConnection::push_upload_data_chunk(const uint8_t *data,
 }
 
 int HttpDownstreamConnection::end_upload_data() {
+  signal_write();
+
   if (!downstream_->get_chunked_request()) {
     return 0;
   }
@@ -577,8 +713,6 @@ int HttpDownstreamConnection::end_upload_data() {
     output->append("\r\n");
   }
 
-  signal_write();
-
   return 0;
 }
 
@@ -617,6 +751,9 @@ namespace {
 void idle_timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
+
+  // We don't have to check conn->expired_rt() since we restart timer
+  // when connection gets idle.
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, dconn) << "Idle connection timeout";
   }
@@ -637,9 +774,13 @@ void HttpDownstreamConnection::detach_downstream(Downstream *downstream) {
 
   auto &downstreamconf = *worker_->get_downstream_config();
 
-  conn_.rt.repeat = downstreamconf.timeout.idle_read;
   ev_set_cb(&conn_.rt, idle_timeoutcb);
-  ev_timer_again(conn_.loop, &conn_.rt);
+  if (conn_.read_timeout < downstreamconf.timeout.idle_read) {
+    conn_.read_timeout = downstreamconf.timeout.idle_read;
+    conn_.last_read = ev_now(conn_.loop);
+  } else {
+    conn_.again_rt(downstreamconf.timeout.idle_read);
+  }
 
   conn_.wlimit.stopw();
   ev_timer_stop(conn_.loop, &conn_.wt);
@@ -681,10 +822,14 @@ namespace {
 int htp_hdrs_completecb(http_parser *htp) {
   auto downstream = static_cast<Downstream *>(htp->data);
   auto upstream = downstream->get_upstream();
+  auto handler = upstream->get_client_handler();
   const auto &req = downstream->request();
   auto &resp = downstream->response();
   int rv;
 
+  auto config = get_config();
+  auto &loggingconf = config->logging;
+
   resp.http_status = htp->status_code;
   resp.http_major = htp->http_major;
   resp.http_minor = htp->http_minor;
@@ -701,7 +846,37 @@ int htp_hdrs_completecb(http_parser *htp) {
   downstream->set_downstream_addr_group(dconn->get_downstream_addr_group());
   downstream->set_addr(dconn->get_addr());
 
-  if (resp.fs.parse_content_length() != 0) {
+  // Server MUST NOT send Transfer-Encoding with a status code 1xx or
+  // 204.  Also server MUST NOT send Transfer-Encoding with a status
+  // code 200 to a CONNECT request.  Same holds true with
+  // Content-Length.
+  if (resp.http_status == 204) {
+    if (resp.fs.header(http2::HD_TRANSFER_ENCODING)) {
+      return -1;
+    }
+    // Some server send content-length: 0 for 204.  Until they get
+    // fixed, we accept, but ignore it.
+
+    // Calling parse_content_length() detects duplicated
+    // content-length header fields.
+    if (resp.fs.parse_content_length() != 0) {
+      return -1;
+    }
+    if (resp.fs.content_length != 0) {
+      return -1;
+    }
+    if (resp.fs.content_length == 0) {
+      auto cl = resp.fs.header(http2::HD_CONTENT_LENGTH);
+      assert(cl);
+      http2::erase_header(cl);
+    }
+  } else if (resp.http_status / 100 == 1 ||
+             (resp.http_status == 200 && req.method == HTTP_CONNECT)) {
+    if (resp.fs.header(http2::HD_CONTENT_LENGTH) ||
+        resp.fs.header(http2::HD_TRANSFER_ENCODING)) {
+      return -1;
+    }
+  } else if (resp.fs.parse_content_length() != 0) {
     downstream->set_response_state(Downstream::MSG_BAD_HEADER);
     return -1;
   }
@@ -740,6 +915,11 @@ int htp_hdrs_completecb(http_parser *htp) {
     downstream->set_chunked_response(false);
   }
 
+  if (loggingconf.access.write_early && downstream->accesslog_ready()) {
+    handler->write_accesslog(downstream);
+    downstream->set_accesslog_written(true);
+  }
+
   if (upstream->on_downstream_header_complete(downstream) != 0) {
     return -1;
   }
@@ -807,7 +987,6 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
   auto downstream = static_cast<Downstream *>(htp->data);
   auto &resp = downstream->response();
   auto &httpconf = get_config()->http;
-  auto &balloc = downstream->get_block_allocator();
 
   if (ensure_header_field_buffer(downstream, httpconf, len) != 0) {
     return -1;
@@ -820,9 +999,7 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
       if (ensure_max_header_fields(downstream, httpconf) != 0) {
         return -1;
       }
-      auto name = http2::copy_lower(balloc, StringRef{data, len});
-      auto token = http2::lookup_token(name);
-      resp.fs.add_header_token(name, StringRef{}, false, token);
+      resp.fs.alloc_add_header_name(StringRef{data, len});
     }
   } else {
     // trailer part
@@ -835,9 +1012,7 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
         // wrong place or crash if trailer fields are currently empty.
         return -1;
       }
-      auto name = http2::copy_lower(balloc, StringRef{data, len});
-      auto token = http2::lookup_token(name);
-      resp.fs.add_trailer_token(name, StringRef{}, false, token);
+      resp.fs.alloc_add_trailer_name(StringRef{data, len});
     }
   }
   return 0;
@@ -904,7 +1079,7 @@ int htp_msg_completecb(http_parser *htp) {
 } // namespace
 
 namespace {
-http_parser_settings htp_hooks = {
+constexpr http_parser_settings htp_hooks = {
     htp_msg_begincb,     // http_cb on_message_begin;
     nullptr,             // http_data_cb on_url;
     nullptr,             // http_data_cb on_status;
@@ -916,7 +1091,33 @@ http_parser_settings htp_hooks = {
 };
 } // namespace
 
+int HttpDownstreamConnection::write_reuse_first() {
+  int rv;
+
+  if (conn_.tls.ssl) {
+    rv = write_tls();
+  } else {
+    rv = write_clear();
+  }
+
+  if (rv != 0) {
+    return SHRPX_ERR_RETRY;
+  }
+
+  if (conn_.tls.ssl) {
+    on_write_ = &HttpDownstreamConnection::write_tls;
+  } else {
+    on_write_ = &HttpDownstreamConnection::write_clear;
+  }
+
+  reuse_first_write_done_ = true;
+
+  return 0;
+}
+
 int HttpDownstreamConnection::read_clear() {
+  conn_.last_read = ev_now(conn_.loop);
+
   std::array<uint8_t, 16_k> buf;
   int rv;
 
@@ -942,6 +1143,8 @@ int HttpDownstreamConnection::read_clear() {
 }
 
 int HttpDownstreamConnection::write_clear() {
+  conn_.last_read = ev_now(conn_.loop);
+
   auto upstream = downstream_->get_upstream();
   auto input = downstream_->get_request_buf();
 
@@ -957,7 +1160,16 @@ int HttpDownstreamConnection::write_clear() {
     }
 
     if (nwrite < 0) {
-      return nwrite;
+      if (!reuse_first_write_done_) {
+        return nwrite;
+      }
+      // We may have pending data in receive buffer which may contain
+      // part of response body.  So keep reading.  Invoke read event
+      // to get read(2) error just in case.
+      ev_feed_event(conn_.loop, &conn_.rev, EV_READ);
+      on_write_ = &HttpDownstreamConnection::noop;
+      reusable_ = false;
+      break;
     }
 
     input->drain(nwrite);
@@ -979,7 +1191,7 @@ int HttpDownstreamConnection::write_clear() {
 int HttpDownstreamConnection::tls_handshake() {
   ERR_clear_error();
 
-  ev_timer_again(conn_.loop, &conn_.rt);
+  conn_.last_read = ev_now(conn_.loop);
 
   auto rv = conn_.tls_handshake();
   if (rv == SHRPX_ERR_INPROGRESS) {
@@ -987,7 +1199,7 @@ int HttpDownstreamConnection::tls_handshake() {
   }
 
   if (rv < 0) {
-    downstream_failure(addr_);
+    downstream_failure(addr_, raddr_);
 
     return rv;
   }
@@ -997,8 +1209,8 @@ int HttpDownstreamConnection::tls_handshake() {
   }
 
   if (!get_config()->tls.insecure &&
-      ssl::check_cert(conn_.tls.ssl, addr_) != 0) {
-    downstream_failure(addr_);
+      ssl::check_cert(conn_.tls.ssl, addr_, raddr_) != 0) {
+    downstream_failure(addr_, raddr_);
 
     return -1;
   }
@@ -1006,22 +1218,22 @@ int HttpDownstreamConnection::tls_handshake() {
   if (!SSL_session_reused(conn_.tls.ssl)) {
     auto session = SSL_get0_session(conn_.tls.ssl);
     if (session) {
-      ssl::try_cache_tls_session(addr_->tls_session_cache, addr_->addr, session,
+      ssl::try_cache_tls_session(addr_->tls_session_cache, *raddr_, session,
                                  ev_now(conn_.loop));
     }
   }
 
   auto &connect_blocker = addr_->connect_blocker;
 
-  do_signal_write_ = &HttpDownstreamConnection::actual_signal_write;
+  signal_write_ = &HttpDownstreamConnection::actual_signal_write;
 
   connect_blocker->on_success();
 
   ev_set_cb(&conn_.rt, timeoutcb);
   ev_set_cb(&conn_.wt, timeoutcb);
 
-  do_read_ = &HttpDownstreamConnection::read_tls;
-  do_write_ = &HttpDownstreamConnection::write_tls;
+  on_read_ = &HttpDownstreamConnection::read_tls;
+  on_write_ = &HttpDownstreamConnection::write_tls;
 
   // TODO Check negotiated ALPN
 
@@ -1029,6 +1241,8 @@ int HttpDownstreamConnection::tls_handshake() {
 }
 
 int HttpDownstreamConnection::read_tls() {
+  conn_.last_read = ev_now(conn_.loop);
+
   ERR_clear_error();
 
   std::array<uint8_t, 16_k> buf;
@@ -1056,6 +1270,8 @@ int HttpDownstreamConnection::read_tls() {
 }
 
 int HttpDownstreamConnection::write_tls() {
+  conn_.last_read = ev_now(conn_.loop);
+
   ERR_clear_error();
 
   auto upstream = downstream_->get_upstream();
@@ -1073,7 +1289,16 @@ int HttpDownstreamConnection::write_tls() {
     }
 
     if (nwrite < 0) {
-      return nwrite;
+      if (!reuse_first_write_done_) {
+        return nwrite;
+      }
+      // We may have pending data in receive buffer which may contain
+      // part of response body.  So keep reading.  Invoke read event
+      // to get read(2) error just in case.
+      ev_feed_event(conn_.loop, &conn_.rev, EV_READ);
+      on_write_ = &HttpDownstreamConnection::noop;
+      reusable_ = false;
+      break;
     }
 
     input->drain(nwrite);
@@ -1167,10 +1392,10 @@ int HttpDownstreamConnection::connected() {
     conn_.wlimit.stopw();
 
     DCLOG(WARN, this) << "Backend connect failed; addr="
-                      << util::to_numeric_addr(&addr_->addr)
+                      << util::to_numeric_addr(raddr_)
                       << ": errno=" << sock_error;
 
-    downstream_failure(addr_);
+    downstream_failure(addr_, raddr_);
 
     return -1;
   }
@@ -1186,36 +1411,37 @@ int HttpDownstreamConnection::connected() {
   ev_timer_again(conn_.loop, &conn_.wt);
 
   conn_.rlimit.startw();
+  conn_.again_rt();
 
   ev_set_cb(&conn_.wev, writecb);
 
   if (conn_.tls.ssl) {
-    do_read_ = &HttpDownstreamConnection::tls_handshake;
-    do_write_ = &HttpDownstreamConnection::tls_handshake;
+    on_read_ = &HttpDownstreamConnection::tls_handshake;
+    on_write_ = &HttpDownstreamConnection::tls_handshake;
 
     return 0;
   }
 
-  do_signal_write_ = &HttpDownstreamConnection::actual_signal_write;
+  signal_write_ = &HttpDownstreamConnection::actual_signal_write;
 
   connect_blocker->on_success();
 
   ev_set_cb(&conn_.rt, timeoutcb);
   ev_set_cb(&conn_.wt, timeoutcb);
 
-  do_read_ = &HttpDownstreamConnection::read_clear;
-  do_write_ = &HttpDownstreamConnection::write_clear;
+  on_read_ = &HttpDownstreamConnection::read_clear;
+  on_write_ = &HttpDownstreamConnection::write_clear;
 
   return 0;
 }
 
-int HttpDownstreamConnection::on_read() { return do_read_(*this); }
+int HttpDownstreamConnection::on_read() { return on_read_(*this); }
 
-int HttpDownstreamConnection::on_write() { return do_write_(*this); }
+int HttpDownstreamConnection::on_write() { return on_write_(*this); }
 
 void HttpDownstreamConnection::on_upstream_change(Upstream *upstream) {}
 
-void HttpDownstreamConnection::signal_write() { do_signal_write_(*this); }
+void HttpDownstreamConnection::signal_write() { signal_write_(*this); }
 
 int HttpDownstreamConnection::actual_signal_write() {
   ev_feed_event(conn_.loop, &conn_.wev, EV_WRITE);
@@ -1231,6 +1457,10 @@ HttpDownstreamConnection::get_downstream_addr_group() const {
 
 DownstreamAddr *HttpDownstreamConnection::get_addr() const { return addr_; }
 
-bool HttpDownstreamConnection::poolable() const { return !group_->retired; }
+bool HttpDownstreamConnection::poolable() const {
+  return !group_->retired && reusable_;
+}
+
+const Address *HttpDownstreamConnection::get_raddr() const { return raddr_; }
 
 } // namespace shrpx
